#
# Sonar Scanner Python
# Copyright (C) 2011-2026 SonarSource Sàrl
# mailto:info AT sonarsource DOT com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful,
#
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
import logging
import os
import json
from typing import Dict

from pysonar_scanner.configuration.properties import Key, PROPERTIES


def load() -> Dict[Key, str]:
    """
    Load configuration properties from environment variables.

    Returns:
        Dictionary of property keys and their values extracted from environment variables.
    """
    # Extract properties from environment variables using the env_variable_name() method

    # First check for JSON params environment variable
    properties = load_json_env_variables()
    # Extract properties from environment variables using the mapping in Property objects
    properties.update(load_properties_env_variables())

    return properties


def load_json_env_variables():
    properties = {}
    if "SONAR_SCANNER_JSON_PARAMS" in os.environ:
        try:
            json_params = os.environ["SONAR_SCANNER_JSON_PARAMS"]
            json_properties = json.loads(json_params)
            properties.update(json_properties)
        except json.JSONDecodeError as e:
            logging.warning(
                f"The JSON in SONAR_SCANNER_JSON_PARAMS environment variable is invalid. The other environment variables will still be loaded. Error : {e}"
            )
    logging.debug("Loaded environment properties: " + ", ".join(f"{key}={value}" for key, value in properties.items()))
    return properties


def load_properties_env_variables():
    properties = {}
    for prop in PROPERTIES:
        env_var_name = prop.env_variable_name()
        if env_var_name in os.environ:
            properties[prop.name] = os.environ[env_var_name]
            if prop.deprecation_message:
                logging.warning(prop.deprecation_message)
    return properties



#
# Sonar Scanner Python
# Copyright (C) 2011-2026 SonarSource Sàrl
# mailto:info AT sonarsource DOT com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful,
#
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
import logging
from pathlib import Path
from typing import Any

from pysonar_scanner.configuration.cli import CliConfigurationLoader
from pysonar_scanner.configuration.coveragerc_loader import CoverageRCConfigurationLoader
from pysonar_scanner.configuration.pyproject_toml import TomlConfigurationLoader
from pysonar_scanner.configuration.properties import SONAR_PROJECT_KEY, SONAR_TOKEN, SONAR_PROJECT_BASE_DIR, Key
from pysonar_scanner.configuration.properties import PROPERTIES
from pysonar_scanner.configuration import sonar_project_properties, environment_variables, dynamic_defaults_loader

from pysonar_scanner.exceptions import MissingProperty, MissingPropertyException


def get_static_default_properties() -> dict[Key, Any]:
    return {prop.name: prop.default_value for prop in PROPERTIES if prop.default_value is not None}


class ConfigurationLoader:
    @staticmethod
    def load() -> dict[Key, Any]:
        logging.debug("Loading configuration properties...")

        # each property loader is required to return NO default values.
        # E.g. if no property has been set, an empty dict must be returned.
        # Default values should be set through the get_static_default_properties() method
        cli_properties = CliConfigurationLoader.load()
        # CLI properties have a higher priority than properties file,
        # but we need to resolve them first to load the properties file
        base_dir = Path(cli_properties.get(SONAR_PROJECT_BASE_DIR, "."))

        toml_path_property = cli_properties.get("toml-path", ".")
        toml_path = Path(toml_path_property) if "toml-path" in cli_properties else base_dir
        toml_properties = TomlConfigurationLoader.load(toml_path)
        coverage_properties = CoverageRCConfigurationLoader.load_exclusion_properties(base_dir)

        resolved_properties = get_static_default_properties()
        resolved_properties.update(dynamic_defaults_loader.load())
        resolved_properties.update(coverage_properties)
        resolved_properties.update(toml_properties.project_properties)
        resolved_properties.update(sonar_project_properties.load(base_dir))
        resolved_properties.update(toml_properties.sonar_properties)
        resolved_properties.update(environment_variables.load())
        resolved_properties.update(cli_properties)
        return resolved_properties

    @staticmethod
    def check_configuration(config: dict[Key, Any]) -> None:
        missing_keys = []
        if SONAR_TOKEN not in config:
            missing_keys.append(MissingProperty(SONAR_TOKEN, "--token"))

        if SONAR_PROJECT_KEY not in config:
            missing_keys.append(MissingProperty(SONAR_PROJECT_KEY, "--project-key"))

        if len(missing_keys) > 0:
            raise MissingPropertyException.from_missing_keys(*missing_keys)


def get_token(config: dict[Key, Any]) -> str:
    if SONAR_TOKEN not in config:
        raise MissingPropertyException(f'Missing property "{SONAR_TOKEN}"')
    return config[SONAR_TOKEN]



#
# Sonar Scanner Python
# Copyright (C) 2011-2026 SonarSource Sàrl
# mailto:info AT sonarsource DOT com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful,
#
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
import argparse
from typing import Any

from pysonar_scanner.configuration import properties
from pysonar_scanner.exceptions import UnexpectedCliArgument


class PyScannerHelpFormatter(argparse.HelpFormatter):
    recommended_args = {"help", "token", "sonar_project_key"}

    def _format_actions_usage(self, actions, groups):
        filtered_actions = [action for action in actions if action.dest in PyScannerHelpFormatter.recommended_args]
        return super()._format_actions_usage(filtered_actions, groups)


class CliConfigurationLoader:

    @classmethod
    def load(cls) -> dict[str, Any]:
        args, unknown_args = cls.__parse_cli_args()
        config = {}
        for prop in properties.PROPERTIES:
            if prop.cli_getter is not None:
                value = prop.cli_getter(args)
                config[prop.name] = value

        # Handle unknown args starting with '-D'
        for arg in unknown_args:
            if not arg.startswith("-D"):
                raise UnexpectedCliArgument(f"Unexpected argument: {arg}\nRun with --help for more information.")
            key_value = arg[2:].split("=", 1)
            if len(key_value) == 2:
                key, value = key_value
                config[key] = value
            else:
                # If no value is provided, set the key to "true"
                config[arg[2:]] = "true"

        return {k: v for k, v in config.items() if v is not None}

    @classmethod
    def __parse_cli_args(cls) -> tuple[argparse.Namespace, list[str]]:
        parser = cls.__create_parser()
        return parser.parse_known_args()

    @classmethod
    def __create_parser(cls):
        parser = argparse.ArgumentParser(
            description="Sonar scanner CLI for Python",
            epilog="Analysis properties not listed here will also be accepted, as long as they start with the -D prefix.",
            formatter_class=PyScannerHelpFormatter,
        )

        parser.add_argument(
            "-t",
            "--token",
            "--sonar-token",
            "-Dsonar.token",
            type=str,
            help="Token used to authenticate against the SonarQube Server or SonarQube Cloud",
        )
        parser.add_argument(
            "--sonar-project-key",
            "--project-key",
            "-Dsonar.projectKey",
            type=str,
            help="Key of the project that usually corresponds to the project name in SonarQube",
        )
        parser.add_argument(
            "--sonar-project-name",
            "-Dsonar.projectName",
            type=str,
            help="Name of the project in SonarQube",
        )
        parser.add_argument(
            "--sonar-sources",
            "-Dsonar.sources",
            type=str,
            help="The analysis scope for main source code (non-test code) in the project",
        )
        parser.add_argument(
            "--sonar-tests",
            "-Dsonar.tests",
            type=str,
            help="The analysis scope for test source code in the project",
        )

        parser.add_argument(
            "--sonar-project-base-dir",
            "-Dsonar.projectBaseDir",
            type=str,
            help="Directory containing the project to be analyzed. Default is the current directory",
        )
        parser.add_argument(
            "--sonar-external-issues-report-paths",
            "-Dsonar.externalIssuesReportPaths",
            type=str,
            help="Comma-delimited list of paths to generic issue reports",
        )
        parser.add_argument(
            "--sonar-sarif-report-paths",
            "-Dsonar.sarifReportPaths",
            type=str,
            help="Comma-delimited list of paths to SARIF issue reports",
        )
        parser.add_argument(
            "--sonar-links-ci", "-Dsonar.links.ci", type=str, help="The URL of the continuous integration system used"
        )
        parser.add_argument(
            "--sonar-links-homepage", "-Dsonar.links.homepage", type=str, help="The URL of the build project home page"
        )
        parser.add_argument(
            "--sonar-links-issue", "-Dsonar.links.issue", type=str, help="The URL to the issue tracker being used"
        )

        parser.add_argument(
            "--sonar-source-encoding",
            "-Dsonar.sourceEncoding",
            type=str,
            help="Encoding of the source files. For example, UTF-8, MacRoman, Shift_JIS",
        )

        parser.add_argument(
            "--toml-path",
            type=str,
            help="Path to the pyproject.toml file or to the folder containing it. If not provided, it will look in the SONAR_PROJECT_BASE_DIR",
        )

        parser.add_argument(
            "--sonar-project-version",
            "-Dsonar.projectVersion",
            type=str,
            help="Version of the project",
        )

        parser.add_argument(
            "--sonar-project-description",
            "-Dsonar.projectDescription",
            type=str,
            help="Description of the project",
        )

        parser.add_argument(
            "--sonar-python-version",
            "-Dsonar.python.version",
            type=str,
            help="Python version used for the project",
        )

        parser.add_argument(
            "--sonar-modules", "-Dsonar.modules", type=str, help="Comma-delimited list of modules to analyze"
        )

        parser.add_argument(
            "--sonar-python-analysis-parallel",
            "--analysis-in-parallel",
            "-Dsonar.python.analysis.parallel",
            action=argparse.BooleanOptionalAction,
            help="When set to False the analysis will be single threaded",
        )
        parser.add_argument(
            "--sonar-python-analysis-threads",
            "--nr-analysis-threads",
            "-Dsonar.python.analysis.threads",
            type=int,
            help="Set the number of threads to use during analysis. This property is ignored if --sonar-python-analysis-parallel is set to False",
        )

        server_connection_group = parser.add_argument_group("SonarQube Connection")
        server_connection_group.add_argument(
            "--sonar-host-url",
            "-Dsonar.host.url",
            type=str,
            help="SonarQube Server base URL. For example, http://localhost:9000 for a local instance of SonarQube Server",
        )
        server_connection_group.add_argument(
            "--sonar-region",
            "-Dsonar.region",
            type=str,
            choices=["us"],
            help="The region to contact, only for SonarQube Cloud",
        )
        server_connection_group.add_argument(
            "--sonar-organization",
            "-Dsonar.organization",
            type=str,
            help="The key of the organization to which the project belongs",
        )
        server_connection_group.add_argument(
            "--sonar-scanner-cloud-url",
            "-Dsonar.scanner.cloudUrl",
            type=str,
            help="SonarQube Cloud base URL, https://sonarcloud.io for example",
        )
        server_connection_group.add_argument(
            "--sonar-scanner-api-url",
            "-Dsonar.scanner.apiUrl",
            type=str,
            help="Base URL for all REST-compliant API calls, https://api.sonarcloud.io for example",
        )
        server_connection_group.add_argument(
            "--sonar-scanner-connect-timeout",
            "-Dsonar.scanner.connectTimeout",
            type=int,
            help="Time period to establish connections with the server (in seconds)",
        )
        server_connection_group.add_argument(
            "--sonar-scanner-socket-timeout",
            "-Dsonar.scanner.socketTimeout",
            type=int,
            help="Maximum time of inactivity between two data packets when exchanging data with the server (in seconds)",
        )
        server_connection_group.add_argument(
            "--sonar-scanner-response-timeout",
            "-Dsonar.scanner.responseTimeout",
            type=int,
            help="Time period required to process an HTTP call: from sending a request to receiving a response (in seconds)",
        )

        scanner_behavior_group = parser.add_argument_group("Scanner Behavior & Advanced Settings")
        scanner_behavior_group.add_argument(
            "-v",
            "--verbose",
            "--sonar-verbose",
            "-Dsonar.verbose",
            action=argparse.BooleanOptionalAction,
            default=None,
            help="Increase output verbosity",
        )
        scanner_behavior_group.add_argument(
            "--sonar-user-home", "-Dsonar.userHome", type=str, help="Base sonar directory, ~/.sonar by default"
        )
        scanner_behavior_group.add_argument(
            "--sonar-scanner-os",
            "-Dsonar.scanner.os",
            type=str,
            choices=["windows", "linux", "macos", "alpine"],
            help="OS running the scanner",
        )
        scanner_behavior_group.add_argument(
            "--sonar-scanner-arch",
            "-Dsonar.scanner.arch",
            type=str,
            choices=["x64", "aarch64"],
            help="Architecture on which the scanner will be running",
        )
        scanner_behavior_group.add_argument(
            "--sonar-scanner-metadata-filepath",
            "-Dsonar.scanner.metadataFilepath",
            type=str,
            help="Sets the location where the scanner writes the report-task.txt file containing among other things the ceTaskId",
        )
        scanner_behavior_group.add_argument(
            "--sonar-scanner-internal-dump-to-file",
            "-Dsonar.scanner.internal.dumpToFile",
            type=str,
            help="Filename where the input to the scanner engine will be dumped. Useful for debugging",
        )
        scanner_behavior_group.add_argument(
            "--sonar-scanner-internal-sq-version",
            "-Dsonar.scanner.internal.sqVersion",
            type=str,
            help="Emulate the result of the call to get SQ server version.  Useful for debugging with --sonar-scanner-internal-dump-to-file",
        )
        scanner_behavior_group.add_argument(
            "--sonar-scm-exclusions-disabled",
            action=argparse.BooleanOptionalAction,
            help="Defines whether files ignored by the SCM, e.g., files listed in .gitignore, will be excluded from the analysis or not",
        )
        scanner_behavior_group.add_argument(
            "-Dsonar.scm.exclusions.disabled",
            type=bool,
            help="Equivalent to --sonar-scm-exclusions-disabled",
        )
        scanner_behavior_group.add_argument(
            "--sonar-filesize-limit",
            "-Dsonar.filesize.limit",
            type=int,
            help="Sets the limit in MB for files to be discarded from the analysis scope if the size is greater than specified",
        )
        scanner_behavior_group.add_argument(
            "--sonar-cpd-py-minimum-tokens",
            "-Dsonar.cpd.py.minimumTokens",
            "--sonar-cpd-python-minimum-tokens",
            "-Dsonar.cpd.python.minimumTokens",
            type=int,
            help="Minimum number of tokens to be considered as a duplicated block of code",
        )
        scanner_behavior_group.add_argument(
            "--sonar-cpd-py-minimum-lines",
            "-Dsonar.cpd.py.minimumLines",
            "--sonar-cpd-python-minimum-lines",
            "-Dsonar.cpd.python.minimumLines",
            type=int,
            help="Minimum number of tokens to be considered as a duplicated block of code",
        )
        scanner_behavior_group.add_argument(
            "--sonar-log-level",
            "-Dsonar.log.level",
            type=str,
            choices=["TRACE", "DEBUG", "INFO", "WARN", "ERROR"],
            help="Log level during the analysis",
        )
        scanner_behavior_group.add_argument(
            "--sonar-qualitygate-wait",
            action=argparse.BooleanOptionalAction,
            help="Forces the analysis step to poll the server instance and wait for the Quality Gate status",
        )
        scanner_behavior_group.add_argument(
            "-Dsonar.qualitygate.wait",
            type=bool,
            help="Equivalent to --sonar-qualitygate-wait",
        )
        scanner_behavior_group.add_argument(
            "--sonar-qualitygate-timeout",
            "-Dsonar.qualitygate.timeout",
            type=int,
            help="The number of seconds that the scanner should wait for a report to be processed",
        )
        scanner_behavior_group.add_argument(
            "--sonar-build-string",
            "-Dsonar.buildString",
            type=str,
            help="The string passed with this property will be stored with the analysis and available in the results of api/project_analyses/search, thus allowing you to later identify a specific analysis and obtain its key for use with api/new_code_periods/set on the SPECIFIC_ANALYSIS type",
        )
        scanner_behavior_group.add_argument(
            "--sonar-working-directory",
            "-Dsonar.working.directory",
            type=str,
            help="Path to the working directory used by the Sonar scanner during a project analysis to store temporary data",
        )
        scanner_behavior_group.add_argument(
            "--sonar-scm-force-reload-all",
            action=argparse.BooleanOptionalAction,
            help="Set this property to true to load blame information for all files, which may significantly increase analysis duration",
        )
        scanner_behavior_group.add_argument(
            "-Dsonar.scm.forceReloadAll",
            type=bool,
            help="Equivalent to --sonar-scm-force-reload-all",
        )
        scanner_behavior_group.add_argument(
            "--sonar-python-skip-unchanged",
            action=argparse.BooleanOptionalAction,
            help="Override the SonarQube configuration of skipping or not the analysis of unchanged Python files",
        )
        scanner_behavior_group.add_argument(
            "--dry-run",
            action=argparse.BooleanOptionalAction,
            default=None,
            help="Enable dry-run mode to validate configuration without connecting to SonarQube server or submitting analysis. See DRY_RUN_MODE.md for details",
        )

        jvm_group = parser.add_argument_group("JVM Settings")
        jvm_group.add_argument(
            "--skip-jre-provisioning",
            "-Dsonar.scanner.skipJreProvisioning",
            action="store_true",
            default=None,
            help="If provided, the provisioning of the JRE will be skipped",
        )
        jvm_group.add_argument(
            "--sonar-scanner-java-exe-path",
            "-Dsonar.scanner.javaExePath",
            type=str,
            help="If defined, the scanner engine will be run with this JRE",
        )
        jvm_group.add_argument(
            "--sonar-scanner-java-opts",
            "-Dsonar.scanner.javaOpts",
            type=str,
            help="Arguments provided to the JVM when running the scanner",
        )
        jvm_group.add_argument(
            "--sonar-scanner-java-heap-size",
            "--java-heap-size",
            "-Dsonar.scanner.javaHeapSize",
            type=str,
            help="Arguments specifies the heap size provided to the JVM when running the scanner",
        )

        truststore_group = parser.add_argument_group("Truststore arguments")
        truststore_group.add_argument(
            "--sonar-scanner-truststore-path",
            "-Dsonar.scanner.truststorePath",
            type=str,
            help="Path to the keystore containing trusted server certificates, used by the Scanner in addition to OS and the built-in certificates",
        )
        truststore_group.add_argument(
            "--sonar-scanner-truststore-password",
            "-Dsonar.scanner.truststorePassword",
            type=str,
            help="Password to access the truststore",
        )

        keystore_group = parser.add_argument_group("Keystore arguments")
        keystore_group.add_argument(
            "--sonar-scanner-keystore-path",
            "-Dsonar.scanner.keystorePath",
            type=str,
            help="Path to the keystore containing the client certificates used by the scanner. By default, <sonar.userHome>/ssl/keystore.p12",
        )
        keystore_group.add_argument(
            "--sonar-scanner-keystore-password",
            "-Dsonar.scanner.keystorePassword",
            type=str,
            help="Password to access the keystore",
        )

        proxy_group = parser.add_argument_group("Proxy arguments")
        proxy_group.add_argument("--sonar-scanner-proxy-host", "-Dsonar.scanner.proxyHost", type=str, help="Proxy host")
        proxy_group.add_argument("--sonar-scanner-proxy-port", "-Dsonar.scanner.proxyPort", type=int, help="Proxy port")
        proxy_group.add_argument("--sonar-scanner-proxy-user", "-Dsonar.scanner.proxyUser", type=str, help="Proxy user")
        proxy_group.add_argument(
            "--sonar-scanner-proxy-password", "-Dsonar.scanner.proxyPassword", type=str, help="Proxy password"
        )

        version_control_group = parser.add_argument_group("Version control arguments")
        version_control_group.add_argument(
            "--sonar-links-scm",
            "-Dsonar.links.scm",
            type=str,
            help="The URL of the build project source code repository",
        )
        version_control_group.add_argument(
            "--sonar-branch-name", "-Dsonar.branch.name", type=str, help="Name of the branch being analyzed"
        )
        version_control_group.add_argument(
            "--sonar-pullrequest-key",
            "-Dsonar.pullrequest.key",
            type=str,
            help="Key of the pull request being analyzed",
        )
        version_control_group.add_argument(
            "--sonar-pullrequest-branch",
            "-Dsonar.pullrequest.branch",
            type=str,
            help="Branch of the pull request being analyzed",
        )
        version_control_group.add_argument(
            "--sonar-pullrequest-base",
            "-Dsonar.pullrequest.base",
            type=str,
            help="Base branch of the pull request being analyzed",
        )
        version_control_group.add_argument(
            "--sonar-newcode-reference-branch",
            "-Dsonar.newCode.referenceBranch",
            type=str,
            help="Reference branch for new code definition",
        )
        version_control_group.add_argument(
            "--sonar-scm-revision",
            "-Dsonar.scm.revision",
            type=str,
            help="Overrides the revision, for instance, the Git sha1, displayed in analysis results",
        )

        reports_group = parser.add_argument_group("3rd party reports arguments")
        reports_group.add_argument(
            "--sonar-python-pylint-report-path",
            "--pylint-report-path",
            "-Dsonar.python.pylint.reportPath",
            type=str,
            help="Path to third-parties issues report file for pylint",
        )
        reports_group.add_argument(
            "--sonar-python-coverage-report-paths",
            "--coverage-report-paths",
            "-Dsonar.python.coverage.reportPaths",
            type=str,
            help="Comma-delimited list of paths to coverage reports in the Cobertura XML format.",
        )
        reports_group.add_argument(
            "--sonar-coverage-exclusions",
            "--sonar.coverage.exclusions",
            "-Dsonar.coverage.exclusions",
            type=str,
            help="Defines the source files to be excluded from the code coverage analysis.",
        )
        reports_group.add_argument(
            "-Dsonar.python.skipUnchanged",
            type=bool,
            help="Equivalent to --sonar-python-skip-unchanged",
        )
        reports_group.add_argument(
            "--sonar-python-xunit-report-path",
            "--xunit-report-path",
            "-Dsonar.python.xunit.reportPath",
            type=str,
            help="Path to the report of test execution, relative to project's root",
        )
        reports_group.add_argument(
            "--sonar-python-xunit-skip-details",
            "--xunit-skip-details",
            action=argparse.BooleanOptionalAction,
            help="When enabled, the test execution statistics is provided only on project level",
        )
        reports_group.add_argument(
            "-Dsonar.python.xunit.skipDetails",
            type=bool,
            help="Equivalent to -Dsonar.python.xunit.skipDetails",
        )
        reports_group.add_argument(
            "--sonar-python-mypy-report-paths",
            "--mypy-report-paths",
            "-Dsonar.python.mypy.reportPaths",
            type=str,
            help="Comma-separated mypy report paths, relative to project's root",
        )
        reports_group.add_argument(
            "--sonar-python-bandit-report-paths",
            "--bandit-report-paths",
            "-Dsonar.python.bandit.reportPaths",
            type=str,
            help="Comma-separated bandit report paths, relative to project's root",
        )
        reports_group.add_argument(
            "--sonar-python-flake8-report-paths",
            "--flake8-report-paths",
            "-Dsonar.python.flake8.reportPaths",
            type=str,
            help="Comma-separated flake8 report paths, relative to project's root",
        )
        reports_group.add_argument(
            "--sonar-python-ruff-report-paths",
            "--ruff-report-paths",
            "-Dsonar.python.ruff.reportPaths",
            type=str,
            help="Comma-separated ruff report paths, relative to project's root",
        )

        return parser



#
# Sonar Scanner Python
# Copyright (C) 2011-2026 SonarSource Sàrl
# mailto:info AT sonarsource DOT com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful,
#
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
import logging
import pathlib
import shutil
import tarfile
import zipfile
from dataclasses import dataclass
from typing import Any, Optional

from pysonar_scanner import utils
from pysonar_scanner.api import JRE, SonarQubeApi
from pysonar_scanner.cache import Cache
from pysonar_scanner.exceptions import (
    ChecksumException,
    NoJreAvailableException,
    UnsupportedArchiveFormat,
)
from pysonar_scanner.exceptions import JreProvisioningException
from pysonar_scanner.configuration.properties import (
    SONAR_SCANNER_JAVA_EXE_PATH,
    SONAR_SCANNER_SKIP_JRE_PROVISIONING,
    SONAR_SCANNER_OS,
    Key,
    SONAR_SCANNER_ARCH,
)


@dataclass(frozen=True)
class JREResolvedPath:
    path: pathlib.Path

    @staticmethod
    def from_string(path: str) -> "JREResolvedPath":
        if not path:
            raise ValueError("JRE path cannot be empty")
        return JREResolvedPath(pathlib.Path(path))


class JREProvisioner:
    def __init__(
        self, api: SonarQubeApi, cache: Cache, sonar_scanner_os: utils.OsStr, sonar_scanner_arch: utils.ArchStr
    ):
        self.api = api
        self.cache = cache
        self.sonar_scanner_os = sonar_scanner_os
        self.sonar_scanner_arch = sonar_scanner_arch

    def provision(self) -> JREResolvedPath:
        jre, resolved_path = self.__attempt_provisioning_jre_with_retry()
        return self.__unpack_jre(jre, resolved_path)

    def __attempt_provisioning_jre_with_retry(self) -> tuple[JRE, pathlib.Path]:
        jre_and_resolved_path = self.__attempt_provisioning_jre()
        if jre_and_resolved_path is None:
            logging.warning("Something went wrong while provisionning the JRE. Retrying...")
            jre_and_resolved_path = self.__attempt_provisioning_jre()
        if jre_and_resolved_path is None:
            raise ChecksumException.create("JRE")

        return jre_and_resolved_path

    def __attempt_provisioning_jre(self) -> Optional[tuple[JRE, pathlib.Path]]:
        jre = self.__get_available_jre()

        jre_path = self.__get_jre_from_cache(jre)
        if jre_path is not None:
            return (jre, jre_path)

        jre_path = self.__download_jre(jre)
        return (jre, jre_path) if jre_path is not None else None

    def __get_available_jre(self) -> JRE:
        jres = self.api.get_analysis_jres(os=self.sonar_scanner_os, arch=self.sonar_scanner_arch)
        if len(jres) == 0:
            raise NoJreAvailableException(
                f"No JREs are available for {self.sonar_scanner_os} and {self.sonar_scanner_arch}"
            )
        return jres[0]

    def __get_jre_from_cache(self, jre: JRE) -> Optional[pathlib.Path]:
        cache_file = self.cache.get_file(jre.filename, jre.sha256)
        return cache_file.filepath if cache_file.is_valid() else None

    def __download_jre(self, jre: JRE) -> Optional[pathlib.Path]:
        cache_file = self.cache.get_file(jre.filename, jre.sha256)
        cache_file.filepath.unlink(missing_ok=True)

        with cache_file.open(mode="wb") as f:
            if jre.download_url is not None:
                self.api.download_file_from_url(jre.download_url, f)
            elif jre.id is not None:
                self.api.download_analysis_jre(jre.id, f)
            else:
                raise JreProvisioningException(
                    "Failed to download the JRE using SonarQube. If this problem persists, you can use the option --sonar-scanner-java-exe-path to use your own local JRE."
                )

        return cache_file.filepath if cache_file.is_valid() else None

    def __unpack_jre(self, jre: JRE, file_path: pathlib.Path) -> JREResolvedPath:
        unzip_dir = self.__prepare_unzip_dir(file_path)
        self.__extract_jre(file_path, unzip_dir)
        return JREResolvedPath(unzip_dir / jre.java_path)

    def __prepare_unzip_dir(self, file_path: pathlib.Path) -> pathlib.Path:
        unzip_dir = self.cache.get_file_path(f"{file_path.name}_unzip")
        try:
            if unzip_dir.exists():
                shutil.rmtree(unzip_dir)
            unzip_dir.mkdir(parents=True)
            return unzip_dir
        except OSError as e:
            raise JreProvisioningException(f"Failed to prepare unzip directory: {unzip_dir}") from e

    def __extract_jre(self, file_path: pathlib.Path, unzip_dir: pathlib.Path):
        if file_path.suffix == ".zip":
            with zipfile.ZipFile(file_path, "r") as zip_ref:
                zip_ref.extractall(unzip_dir)
        elif file_path.suffix in [".gz", ".tgz"]:
            utils.extract_tar(file_path, unzip_dir)
        else:
            raise UnsupportedArchiveFormat(
                f"Received JRE is packaged as an unsupported archive format: {file_path.suffix}"
            )


@dataclass(frozen=True)
class JREResolverConfiguration:
    sonar_scanner_java_exe_path: Optional[str]
    sonar_scanner_skip_jre_provisioning: bool
    sonar_scanner_os: Optional[str]

    @staticmethod
    def from_dict(config_dict: dict[Key, Any]) -> "JREResolverConfiguration":
        return JREResolverConfiguration(
            sonar_scanner_java_exe_path=config_dict.get(SONAR_SCANNER_JAVA_EXE_PATH, None),
            sonar_scanner_skip_jre_provisioning=config_dict.get(SONAR_SCANNER_SKIP_JRE_PROVISIONING, False),
            sonar_scanner_os=config_dict.get(SONAR_SCANNER_OS, None),
        )


class JREResolver:
    def __init__(self, configuration: JREResolverConfiguration, jre_provisioner: JREProvisioner):
        self.configuration = configuration
        self.jre_provisioner = jre_provisioner

    def resolve_jre(self) -> JREResolvedPath:
        exe_suffix = ".exe" if self.configuration.sonar_scanner_os == "windows" else ""
        if self.configuration.sonar_scanner_java_exe_path:
            return JREResolvedPath(pathlib.Path(self.configuration.sonar_scanner_java_exe_path))
        if not self.configuration.sonar_scanner_skip_jre_provisioning:
            return self.__provision_jre()
        java_path = pathlib.Path(f"java{exe_suffix}")
        return JREResolvedPath(java_path)

    def __provision_jre(self) -> JREResolvedPath:
        return self.jre_provisioner.provision()



#
# Sonar Scanner Python
# Copyright (C) 2011-2026 SonarSource Sàrl
# mailto:info AT sonarsource DOT com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful,
#
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
import configparser
import logging
import pathlib
from typing import Any


class CoverageRCConfigurationLoader:

    @staticmethod
    def load_exclusion_properties(base_dir: pathlib.Path) -> dict[str, str]:
        config_file_path = base_dir / ".coveragerc"
        coverage_properties = CoverageRCConfigurationLoader.__read_config(config_file_path)
        if len(coverage_properties) == 0:
            return {}
        translated_exclusions = CoverageRCConfigurationLoader.__extract_coverage_exclusion_patterns(
            config_file_path, coverage_properties
        )
        if len(translated_exclusions) == 0:
            return {}

        return {"sonar.coverage.exclusions": ", ".join(translated_exclusions)}

    @staticmethod
    def __read_config(config_file_path: pathlib.Path) -> dict[str, Any]:
        config_dict: dict[str, Any] = {}
        if not config_file_path.exists():
            logging.debug(f"Coverage file not found: {config_file_path}")
            return config_dict

        try:
            config_parser = configparser.ConfigParser()
            config_parser.read(config_file_path)
            for section in config_parser.sections():
                section_values = {}
                for key, value in config_parser.items(section):
                    section_values[key] = value
                config_dict[section] = section_values
        except Exception as e:
            logging.debug(f"Error decoding coverage file {config_file_path}: {e}")
        return config_dict

    @staticmethod
    def __extract_coverage_exclusion_patterns(
        config_file_path: pathlib.Path, coverage_properties: dict[str, Any]
    ) -> list[str]:
        if "run" not in coverage_properties or "omit" not in coverage_properties["run"]:
            logging.debug(f"Coverage file {config_file_path} has no exclusion properties")
            return []

        omit_exclusions = coverage_properties["run"]["omit"]
        return [patterns.strip() for patterns in omit_exclusions.splitlines() if patterns.strip()]



#
# Sonar Scanner Python
# Copyright (C) 2011-2026 SonarSource Sàrl
# mailto:info AT sonarsource DOT com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful,
#
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

import logging
from typing import Any
from pysonar_scanner import app_logging
from pysonar_scanner import cache
from pysonar_scanner import exceptions
from pysonar_scanner.api import get_base_urls, SonarQubeApi, BaseUrls, MIN_SUPPORTED_SQ_VERSION
from pysonar_scanner.configuration import configuration_loader
from pysonar_scanner.configuration.configuration_loader import ConfigurationLoader
from pysonar_scanner.configuration.properties import (
    SONAR_VERBOSE,
    SONAR_HOST_URL,
    SONAR_SCANNER_API_BASE_URL,
    SONAR_SCANNER_SONARCLOUD_URL,
    SONAR_SCANNER_PROXY_PORT,
    SONAR_SCANNER_JAVA_EXE_PATH,
    SONAR_SCANNER_OS,
    SONAR_SCANNER_ARCH,
    SONAR_SCANNER_DRY_RUN,
    SONAR_PROJECT_BASE_DIR,
    SONAR_PYTHON_COVERAGE_REPORT_PATHS,
)
from pysonar_scanner.exceptions import SQTooOldException
from pysonar_scanner.jre import JREResolvedPath, JREProvisioner, JREResolver, JREResolverConfiguration
from pysonar_scanner.scannerengine import ScannerEngine, ScannerEngineProvisioner
from pysonar_scanner.dry_run_reporter import DryRunReporter, CoverageReportValidator, ValidationResult


def main():
    exit(scan())


def scan():
    try:
        return do_scan()
    except Exception as e:
        return exceptions.log_error(e)


def do_scan():
    app_logging.setup()
    logging.info(
        "Enhance your workflow: Pair pysonar with SonarQube Server per your license or SonarQube Cloud for deeper analysis, and try SonarQube-IDE in your favourite IDE."
    )
    logging.info("Starting Pysonar, the Sonar scanner CLI for Python")
    config = ConfigurationLoader.load()
    set_logging_options(config)

    if config.get(SONAR_SCANNER_DRY_RUN, False):
        return run_dry_run(config)

    ConfigurationLoader.check_configuration(config)

    api = build_api(config)
    check_version(api)
    update_config_with_api_urls(config, api.base_urls)
    logging.debug(f"Final loaded configuration: {config}")

    cache_manager = cache.get_cache(config)
    scanner = create_scanner_engine(api, cache_manager, config)

    logging.info("Starting the analysis...")
    return scanner.run(config)


def set_logging_options(config):
    app_logging.configure_logging_level(verbose=config.get(SONAR_VERBOSE, False))


def build_api(config: dict[str, Any]) -> SonarQubeApi:
    token = configuration_loader.get_token(config)
    base_urls = get_base_urls(config)
    return SonarQubeApi(base_urls, token)


def check_version(api: SonarQubeApi):
    if api.is_sonar_qube_cloud():
        logging.debug(f"SonarQube Cloud url: {api.base_urls.base_url}")
        return
    version = api.get_analysis_version()
    logging.debug(f"SonarQube url: {api.base_urls.base_url}")

    if not version.does_support_bootstrapping():
        raise SQTooOldException(
            f"This scanner only supports SonarQube versions >= {MIN_SUPPORTED_SQ_VERSION}. \n"
            f"The server at {api.base_urls.base_url} is on version {version}\n"
            "Please either upgrade your SonarQube server or use the Sonar Scanner CLI (see https://docs.sonarsource.com/sonarqube-server/latest/analyzing-source-code/scanners/sonarscanner/)."
        )


def update_config_with_api_urls(config, base_urls: BaseUrls):
    config[SONAR_HOST_URL] = base_urls.base_url
    config[SONAR_SCANNER_API_BASE_URL] = base_urls.api_base_url
    if base_urls.is_sonar_qube_cloud:
        config[SONAR_SCANNER_SONARCLOUD_URL] = base_urls.base_url
    config[SONAR_SCANNER_PROXY_PORT] = "443" if base_urls.base_url.startswith("https") else "80"


def create_scanner_engine(api, cache_manager, config):
    jre_path = create_jre(api, cache_manager, config)
    config[SONAR_SCANNER_JAVA_EXE_PATH] = str(jre_path.path)
    logging.debug(f"JRE path: {jre_path.path}")
    scanner_engine_path = ScannerEngineProvisioner(api, cache_manager).provision()
    scanner = ScannerEngine(jre_path, scanner_engine_path)
    return scanner


def create_jre(api, cache, config: dict[str, Any]) -> JREResolvedPath:
    jre_provisioner = JREProvisioner(api, cache, config[SONAR_SCANNER_OS], config[SONAR_SCANNER_ARCH])
    jre_resolver = JREResolver(JREResolverConfiguration.from_dict(config), jre_provisioner)
    return jre_resolver.resolve_jre()


def run_dry_run(config: dict[str, Any]) -> int:
    """
    Run in dry-run mode without connecting to SonarQube server.
    Validates configuration and coverage reports.
    """
    logging.info("Running in DRY RUN mode")
    logging.info("No server connection will be made and no analysis will be submitted")

    DryRunReporter.report_configuration(config)

    coverage_paths = config.get(SONAR_PYTHON_COVERAGE_REPORT_PATHS)
    project_base_dir = config.get(SONAR_PROJECT_BASE_DIR, ".")
    validation_result = CoverageReportValidator.validate_coverage_reports(coverage_paths, project_base_dir)

    return DryRunReporter.report_validation_results(validation_result)



#
# Sonar Scanner Python
# Copyright (C) 2011-2026 SonarSource Sàrl
# mailto:info AT sonarsource DOT com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful,
#
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
import json
import logging
import pathlib
import shlex
from dataclasses import dataclass
from subprocess import PIPE, Popen
from threading import Thread
from typing import IO, Any, Callable, Optional

from pysonar_scanner.api import EngineInfo, SonarQubeApi
from pysonar_scanner.cache import Cache, CacheFile
from pysonar_scanner.configuration.properties import (
    SONAR_SCANNER_JAVA_OPTS,
    SONAR_SCANNER_OPTS,
)
from pysonar_scanner.exceptions import ChecksumException
from pysonar_scanner.jre import JREResolvedPath


@dataclass(frozen=True)
class LogLine:
    level: str
    message: str
    stacktrace: Optional[str] = None

    def get_logging_level(self) -> int:
        if self.level == "ERROR":
            return logging.ERROR
        if self.level == "WARN":
            return logging.WARNING
        if self.level == "INFO":
            return logging.INFO
        if self.level == "DEBUG":
            return logging.DEBUG
        if self.level == "TRACE":
            return logging.DEBUG
        return logging.INFO


def parse_log_line(line: str) -> LogLine:
    try:
        line_json = json.loads(line)
        level = line_json.get("level", "INFO")
        message = line_json.get("message", line)
        stacktrace = line_json.get("stacktrace")
        return LogLine(level=level, message=message, stacktrace=stacktrace)
    except json.JSONDecodeError:
        return LogLine(level="INFO", message=line, stacktrace=None)


def default_log_line_listener(log_line: LogLine):
    logging.log(log_line.get_logging_level(), log_line.message)
    if log_line.stacktrace is not None:
        logging.log(log_line.get_logging_level(), log_line.stacktrace)


class CmdExecutor:
    def __init__(
        self,
        cmd: list[str],
        properties_str: str,
        log_line_listener: Callable[[LogLine], None] = default_log_line_listener,
    ):
        self.cmd = cmd
        self.properties_str = properties_str
        self.log_line_listener = log_line_listener

    def execute(self):
        process = Popen(self.cmd, stdin=PIPE, stdout=PIPE, stderr=PIPE)
        process.stdin.write(self.properties_str.encode())
        process.stdin.close()

        output_thread = Thread(target=self.__log_output, args=(process.stdout,))
        error_thread = Thread(target=self.__log_output, args=(process.stderr,))

        return self.__process_output(output_thread, error_thread, process)

    def __log_output(self, stream: IO[bytes]):
        for line in stream:
            decoded_line = line.decode("utf-8").rstrip()
            log_line = parse_log_line(decoded_line)
            self.log_line_listener(log_line)

    def __process_output(self, output_thread: Thread, error_thread: Thread, process: Popen) -> int:
        output_thread.start()
        error_thread.start()
        process.wait()
        output_thread.join()
        error_thread.join()

        return process.returncode


class ScannerEngineProvisioner:
    def __init__(self, api: SonarQubeApi, cache: Cache):
        self.api = api
        self.cache = cache

    def provision(self) -> pathlib.Path:
        scanner_file = self.__download_and_verify()
        if scanner_file is not None:
            return scanner_file.filepath
        # Retry once in case the checksum failed due to the scanner engine being updated between getting the checksum and downloading the jar
        logging.warning("Something went wrong while downloading the scanner engine. Retrying...")
        scanner_file = self.__download_and_verify()
        if scanner_file is not None:
            return scanner_file.filepath
        else:
            raise ChecksumException.create("scanner engine JAR")

    def __download_and_verify(self) -> Optional[CacheFile]:
        engine_info = self.api.get_analysis_engine()
        cache_file = self.cache.get_file(engine_info.filename, engine_info.sha256)
        if not cache_file.is_valid():
            logging.debug("No valid cached analysis engine jar was found")
            self.__download_scanner_engine(cache_file, engine_info)
        return cache_file if cache_file.is_valid() else None

    def __download_scanner_engine(self, cache_file: CacheFile, engine_info: EngineInfo) -> None:
        with cache_file.open(mode="wb") as f:
            if engine_info.download_url is not None:
                self.api.download_file_from_url(engine_info.download_url, f)
            else:
                self.api.download_analysis_engine(f)


class ScannerEngine:
    def __init__(self, jre_path: JREResolvedPath, scanner_engine_path: pathlib.Path):
        self.jre_path = jre_path
        self.scanner_engine_path = scanner_engine_path

    def run(self, config: dict[str, Any]):
        # Extract Java options if present; they must influence the JVM invocation, not the scanner engine itself
        java_opts = config.get(SONAR_SCANNER_JAVA_OPTS)
        java_opts = config.get(SONAR_SCANNER_OPTS) if not java_opts else java_opts

        cmd = self.__build_command(self.jre_path, self.scanner_engine_path, java_opts)
        logging.debug(f"Command: {cmd}")
        properties_str = self.__config_to_json(config)
        logging.debug(f"Properties: {properties_str}")
        return CmdExecutor(cmd, properties_str).execute()

    def __build_command(
        self,
        jre_path: JREResolvedPath,
        scanner_engine_path: pathlib.Path,
        java_opts: Optional[str] = None,
    ) -> list[str]:
        cmd: list[str] = []
        cmd.append(str(jre_path.path))

        if java_opts:
            cmd.extend(self.__decompose_java_opts(java_opts))

        cmd.append("-jar")
        cmd.append(str(scanner_engine_path))
        return cmd

    def __config_to_json(self, config: dict[str, Any]) -> str:
        # SONAR_SCANNER_JAVA_OPTS are properties that shouldn't be passed to the engine, only to the JVM
        scanner_properties = [
            {"key": k, "value": v}
            for k, v in config.items()
            if k != SONAR_SCANNER_JAVA_OPTS and k != SONAR_SCANNER_OPTS
        ]
        return json.dumps({"scannerProperties": scanner_properties})

    def __decompose_java_opts(self, java_opts: str) -> list[str]:
        return shlex.split(java_opts.strip())



#
# Sonar Scanner Python
# Copyright (C) 2011-2026 SonarSource Sàrl
# mailto:info AT sonarsource DOT com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful,
#
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
import logging
from pathlib import Path
from typing import Dict
import os
from jproperties import Properties


def load(base_dir: Path) -> Dict[str, str]:
    filepath = base_dir / "sonar-project.properties"
    if not os.path.isfile(filepath):
        logging.debug(f"no sonar-project.properties file found at {filepath}")
        return {}

    logging.debug(f"sonar-project.properties loaded from {filepath}")
    props = Properties()
    with open(filepath, "rb") as f:
        props.load(f)

    return {key: value.data for key, value in props.items() if value.data is not None}



#
# Sonar Scanner Python
# Copyright (C) 2011-2026 SonarSource Sàrl
# mailto:info AT sonarsource DOT com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful,
#
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
import logging
import sys


class LevelFilter(logging.Filter):
    def __init__(self, level):
        super().__init__()
        self.__level = level

    def filter(self, record):
        return record.levelno < self.__level


def setup() -> None:
    logger = logging.getLogger()

    non_error_handler = logging.StreamHandler(sys.stdout)
    non_error_handler.setLevel(logging.DEBUG)
    non_error_handler.addFilter(LevelFilter(logging.ERROR))

    error_handler = logging.StreamHandler(sys.stderr)
    error_handler.setLevel(logging.ERROR)

    formatter = logging.Formatter("%(levelname)s: %(message)s")
    non_error_handler.setFormatter(formatter)
    error_handler.setFormatter(formatter)

    logger.addHandler(non_error_handler)
    logger.addHandler(error_handler)


def configure_logging_level(verbose: bool) -> None:
    logging.getLogger().setLevel(logging.DEBUG if verbose else logging.INFO)



#
# Sonar Scanner Python
# Copyright (C) 2011-2026 SonarSource Sàrl
# mailto:info AT sonarsource DOT com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful,
#
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

import logging
import re
import xml.etree.ElementTree as ET
from pathlib import Path
from typing import Any, Optional

from pysonar_scanner.configuration.properties import (
    SONAR_PROJECT_KEY,
    SONAR_ORGANIZATION,
    SONAR_SOURCES,
    SONAR_TESTS,
    SONAR_PYTHON_COVERAGE_REPORT_PATHS,
    SONAR_PROJECT_NAME,
    SONAR_HOST_URL,
)


class DryRunReporter:
    @staticmethod
    def report_configuration(config: dict[str, Any]) -> None:
        logging.info("=" * 80)
        logging.info("DRY RUN MODE - Configuration Report")
        logging.info("=" * 80)

        DryRunReporter._log_section(
            "Project Configuration",
            {
                SONAR_PROJECT_KEY: config.get(SONAR_PROJECT_KEY),
                SONAR_PROJECT_NAME: config.get(SONAR_PROJECT_NAME),
                SONAR_ORGANIZATION: config.get(SONAR_ORGANIZATION, "N/A"),
            },
        )

        DryRunReporter._log_section(
            "Server Configuration",
            {
                SONAR_HOST_URL: config.get(SONAR_HOST_URL, "N/A"),
            },
        )

        DryRunReporter._log_section(
            "Source Configuration",
            {
                SONAR_SOURCES: config.get(SONAR_SOURCES, "N/A"),
                SONAR_TESTS: config.get(SONAR_TESTS, "N/A"),
            },
        )

        DryRunReporter._log_section(
            "Coverage Configuration",
            {
                SONAR_PYTHON_COVERAGE_REPORT_PATHS: config.get(SONAR_PYTHON_COVERAGE_REPORT_PATHS, "N/A"),
            },
        )

    @staticmethod
    def report_validation_results(validation_result: "ValidationResult") -> int:
        logging.info("=" * 80)
        logging.info("DRY RUN MODE - Validation Results")
        logging.info("=" * 80)

        if validation_result.is_valid():
            for info in validation_result.infos:
                logging.info(f"✓ {info}")
            for warning in validation_result.warnings:
                logging.warning(f"• {warning}")
            logging.info("✓ Configuration validation PASSED")
            logging.info("=" * 80)
            return 0
        else:
            logging.warning("✗ Configuration validation FAILED with the following issues:")
            for error in validation_result.errors:
                logging.error(f"✗ {error}")
            for warning in validation_result.warnings:
                logging.warning(f"⚠ {warning}")
            logging.info("=" * 80)
            return 1

    @staticmethod
    def _log_section(title: str, values: dict[str, Any]) -> None:
        logging.info(f"\n{title}:")
        for key, value in values.items():
            formatted_key = DryRunReporter._format_key(key)
            logging.info(f"  {formatted_key}: {value}")

    @staticmethod
    def _format_key(key: str) -> str:
        if key.startswith("sonar."):
            key = key[6:]
        key = key.replace(".", " ").replace("_", " ")
        key = re.sub(r"([a-z])([A-Z])", r"\1 \2", key)
        return key.title()


class ValidationResult:
    def __init__(self):
        self.errors: list[str] = []
        self.warnings: list[str] = []
        self.infos: list[str] = []

    def add_error(self, message: str) -> None:
        self.errors.append(message)

    def add_warning(self, message: str) -> None:
        self.warnings.append(message)

    def add_info(self, message: str) -> None:
        self.infos.append(message)

    def is_valid(self) -> bool:
        return len(self.errors) == 0


class CoverageReportValidator:
    @staticmethod
    def validate_coverage_reports(
        coverage_paths: Optional[str],
        project_base_dir: str,
    ) -> ValidationResult:
        validation_result = ValidationResult()
        if not coverage_paths:
            validation_result.add_warning("No coverage report paths specified")
            return validation_result

        base_path = Path(project_base_dir)
        report_paths = [p.strip() for p in coverage_paths.split(",")]

        for report_path in report_paths:
            CoverageReportValidator._validate_single_report(report_path, base_path, validation_result)

        return validation_result

    @staticmethod
    def _validate_single_report(report_path: str, base_path: Path, validation_result: ValidationResult) -> None:
        # Resolve relative path
        full_path = base_path / report_path if not Path(report_path).is_absolute() else Path(report_path)

        if not full_path.exists():
            validation_result.add_error(f"Coverage report not found: {report_path} (resolved to {full_path})")
            return

        if not full_path.is_file():
            validation_result.add_error(f"Coverage report is not a file: {report_path} (resolved to {full_path})")
            return

        try:
            with open(full_path, "r", encoding="utf-8") as f:
                tree = ET.parse(f)
                root = tree.getroot()
                if root.tag != "coverage":
                    validation_result.add_warning(
                        f"Coverage report root element is '{root.tag}', expected 'coverage' (Cobertura format)"
                    )
                else:
                    validation_result.add_info(f"Coverage report check passed: {report_path}")
        except PermissionError:
            validation_result.add_error(f"Coverage report is not readable (permission denied): {report_path}")
        except UnicodeDecodeError:
            validation_result.add_warning(
                f"Coverage report may not be text-based (is it in binary format?): {report_path}"
            )
        except ET.ParseError as e:
            validation_result.add_error(
                f"Coverage report is not valid XML (Cobertura format): {report_path} (Parse error: {str(e)})"
            )
        except Exception as e:
            validation_result.add_error(f"Error validating coverage report format: {report_path} (Error: {str(e)})")



#
# Sonar Scanner Python
# Copyright (C) 2011-2026 SonarSource Sàrl
# mailto:info AT sonarsource DOT com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful,
#
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#


from dataclasses import dataclass
import logging

EXCEPTION_RETURN_CODE = 1


@dataclass
class MissingProperty:
    property: str
    cli_arg: str


class MissingPropertyException(Exception):
    @staticmethod
    def from_missing_keys(*properties: MissingProperty) -> "MissingPropertyException":
        missing_properties = ", ".join([f"{prop.property} ({prop.cli_arg})" for prop in properties])
        fix_message = (
            "You can provide these properties using one of the following methods:\n"
            "- Command line arguments (e.g., --sonar-project-key=myproject)\n"
            "- Analysis parameters with -D prefix (e.g., -Dsonar.projectKey=myproject)\n"
            "- Environment variables (e.g., SONAR_PROJECTKEY=myproject)\n"
            "- Properties file (sonar-project.properties)\n"
            "- Project configuration files (e.g., build.gradle, pom.xml)"
        )
        return MissingPropertyException(f"Missing required properties: {missing_properties}\n\n{fix_message}")


class SonarQubeApiException(Exception):
    pass


class SonarQubeApiUnauthroizedException(SonarQubeApiException):
    @staticmethod
    def create_default(server_url: str) -> "SonarQubeApiUnauthroizedException":
        return SonarQubeApiUnauthroizedException(
            f'The provided token is invalid for the server at "{server_url}". Please check that both the token and the server URL are correct.'
        )


class SQTooOldException(Exception):
    pass


class InconsistentConfiguration(Exception):
    pass


class ChecksumException(Exception):
    @staticmethod
    def create(what: str) -> "ChecksumException":
        return ChecksumException(f"Checksum mismatch. The downloaded {what} is corrupted.")


class UnexpectedCliArgument(Exception):
    pass


class JreProvisioningException(Exception):
    pass


class NoJreAvailableException(JreProvisioningException):
    pass


class UnsupportedArchiveFormat(JreProvisioningException):
    pass


def log_error(e: Exception):
    logger = logging.getLogger()
    is_debug_level = logger.getEffectiveLevel() <= logging.DEBUG

    if is_debug_level:
        logger.error("The following exception occured while running the analysis", exc_info=True)
    else:
        logger.error(str(e), exc_info=False)
        logger.info("For more details, please enable debug logging by passing the --verbose option.")

    return EXCEPTION_RETURN_CODE
